//
// Created by jiapeiyuan on 4/24/21.
//
#include <string>
#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include "client.h"

cv::Mat resize_image(const cv::Mat& img, int tar_w, int tar_h){
    cv::Mat sample_temp;
    cv::Mat sample_resized;

    float img_w = img.cols;
    float img_h = img.rows;
    int new_w = int(img_w * std::min((float)tar_w / img_w, (float)tar_h / img_h));
    int new_h = int(img_h * std::min((float)tar_w / img_w, (float)tar_h / img_h));
    cv::resize(img, sample_temp, cv::Size(new_w,new_h), CV_INTER_CUBIC);
    sample_resized = cv::Mat(tar_h, tar_w, CV_8UC3, cv::Scalar(114, 114, 114));
    sample_temp.copyTo(
            sample_resized(
                    cv::Range(((float)tar_h - new_h) / 2,
                                ((float)tar_h - new_h) / 2 + new_h),
                    cv::Range(((float)tar_w - new_w) / 2,
                                ((float)tar_w - new_w) / 2 + new_w)));
    return sample_resized;

}


void prepare_mlu_data(string image, int batch, void** in_mlu, void** out_mlu){

    int in_h = 384;
    int in_w = 640;
    int in_c = 3;
    int out_h = 1;
    int out_w = 1;
    int out_c = 7232;
    int in_size = in_h * in_w * in_c * sizeof(float);
    int in_size_mlu = in_size /2;
    int out_size = out_h * out_w * out_c * sizeof(float);
    int out_size_mlu = out_size /2;

    cv::Mat img = cv::imread(image, -1);
    cv::Mat resize_img = resize_image(img, in_w, in_h);

    cv::Mat float_img;
    resize_img.convertTo(float_img, CV_32FC3);

    cv::Mat rgb_img;
    cv::cvtColor(float_img, rgb_img, cv::COLOR_BGR2RGB);
    cv::Mat normalized_img;
    normalized_img = rgb_img/255.f;

    void* in_cpu = malloc(in_size*batch);
    void* in_trans_cpu = malloc(in_size_mlu*batch);

    for(int i=0; i<batch; i++){
        memcpy((char*)in_cpu + i*in_size, normalized_img.data, in_size);
    }

    int dim_values[4] = {batch, in_h, in_w, in_c};
    int dim_order[4] = {0, 1, 2, 3};
    cnrtTransOrderAndCast(in_cpu, CNRT_FLOAT32, in_trans_cpu, CNRT_FLOAT16, nullptr, 4, dim_values, dim_order);

    cnrtMalloc(in_mlu, in_size_mlu*batch);
    cnrtMalloc(out_mlu, out_size_mlu*batch);
    cnrtMemcpy(in_mlu[0], in_trans_cpu, in_size_mlu*batch, CNRT_MEM_TRANS_DIR_HOST2DEV);
}

void post(void* out_mlu, int batch){
    int out_h = 1;
    int out_w = 1;
    int out_c = 7232;
    int out_size = out_h * out_w * out_c * sizeof(float);
    int trans_out_size = out_size/2;

    void* out_cpu = malloc(out_size*batch);
    void* out_trans_cpu = malloc(trans_out_size*batch);

    int dim_values[4] = {batch, out_h, out_w, out_c};
    int dim_order[4] = {0, 1, 2, 3};
    cnrtMemcpy(out_trans_cpu, out_mlu, trans_out_size*batch, CNRT_MEM_TRANS_DIR_DEV2HOST);
    cnrtTransOrderAndCast(out_trans_cpu, CNRT_FLOAT16, out_cpu, CNRT_FLOAT32, nullptr, 4, dim_values, dim_order);

    for(int i=0; i<batch; i++){
        printf("%d: %f\n", i, *(float*)((char*)out_cpu+ i*out_size));
    }
}

int main(int argc, char* argv[])
{
    InferClient client = InferClient(4, "testcase/input/model/yolov5s_int8_1b_1c.cambricon");
    string image = "testcase/input/image/imagenet/0.jpg";
//    string image = "testcase/input/image/ial_image/32848.jpg";

    int batch = 32;
    void* in_mlu, *out_mlu;
    prepare_mlu_data(image, batch, &in_mlu, &out_mlu);
    client.process(&in_mlu, &out_mlu, batch);

    post(out_mlu, batch);
}