//
// Created by jiapeiyuan on 4/25/21.
//

#ifndef CS_DEMO_CLIENT_H
#define CS_DEMO_CLIENT_H

#include <string>
#include <unistd.h>
#include "blockqueue.h"
#include "cnrt.h"

using namespace std;

typedef struct inferworkinfo{
    int id;
    BlockQueue<std::pair<void**, void**>> *qin;
    BlockQueue<int> *qout;
}InferWorkInfo;

#define output_num 20
struct offlineInfo{
    cnrtModel_t model;
    cnrtFunction_t function;

    int64_t* inputSizeArray;
    int64_t* outputSizeArray;
    int* input_dims[4];
    int input_dims_num[4];
    int* output_dims[output_num];
    int output_dims_num[output_num];

    int inputNum, outputNum;
    unsigned int input_n, input_c, input_h, input_w;
    unsigned int output_n[output_num], output_c[output_num], output_h[output_num], output_w[output_num];
    int out_count[output_num];
};

class InferClient{
public:
    InferClient(int server_num, string offlinemodel, int device_id=0);
    void process(void** mlu_in, void** mlu_out, int batch);

private:
    void init();
    void CreateServer(int i);
    bool init_device();
    bool parse_input_para();
    bool parse_output_para();
    void infer_cur_data(void** mlu_input, void** mlu_output);
    void fetch_result();
    int get_processed_batch();
    int get_fetched_batch();
    void* get_mlu_in_ptr(void* mlu_in);
    void* get_mlu_out_ptr(void* mlu_out);
    bool is_server_availed();

private:
    int server_num_;
    vector<InferWorkInfo*> servers_;
    vector<vector<void*>> mlu_input_;
    vector<vector<void*>> mlu_output_;
    
    int process_idx_;
    int fetch_idx_;

    int rest_;

    int device_id_;
    offlineInfo offline_info_;
    string offlinemodel_;

    int processed_batch_;
    int fetched_batch_;
};

#endif //CS_DEMO_CLIENT_H
