//
// Created by jiapeiyuan on 4/24/21.
//

#include "client.h"
#include "server.h"


InferClient :: InferClient(int server_num, string offlinemodel, int device_id) {
    offlinemodel_ = offlinemodel;
    device_id_ = device_id;
    server_num_ = server_num;
    mlu_input_.reserve(server_num);
    mlu_output_.reserve(server_num);
    for(int i=0; i<server_num; i++){
        mlu_input_[i].reserve(1);
        mlu_input_[i].push_back((void*)malloc(sizeof(void*)));
        mlu_output_[i].reserve(1);
        mlu_output_[i].push_back((void*)malloc(sizeof(void*)));
    }

    init();
    cnrtInit(0);
    parse_input_para();
    parse_output_para();
    init_device();

    for(int i=0; i<server_num_; i++){
        CreateServer(i);
    }
}

void InferClient::init(){
    process_idx_ = 0;
    fetch_idx_ = 0;
    processed_batch_ = 0;
    fetched_batch_ = 0;
    rest_ = server_num_;
}


void InferClient::CreateServer(int i){
    InferWorkInfo *task = new InferWorkInfo();
    BlockQueue<std::pair<void**, void**>>* in = new BlockQueue<std::pair<void**, void**>>(1);
    BlockQueue<int>* out = new BlockQueue<int>(1);
    task->id = i;
    task->qin = in;
    task->qout = out;
    servers_.push_back(task);

    new InferServer(i, offlinemodel_.c_str(), in, out);
}

bool InferClient :: init_device() {
    cnrtInit(0);
    unsigned dev_num;
    cnrtGetDeviceCount(&dev_num);
    cnrtDev_t dev;
    CNRT_CHECK(cnrtGetDeviceHandle(&dev, device_id_));
    CNRT_CHECK(cnrtSetCurrentDevice(dev));
    cnrtSetCurrentChannel((cnrtChannelType_t)CNRT_CHANNEL_TYPE_NONE);
    return true;
}


bool InferClient ::parse_input_para() {
    cnrtInit(0);
    cnrtLoadModel(&offline_info_.model, offlinemodel_.c_str());

    cnrtCreateFunction(&offline_info_.function);
    cnrtExtractFunction(&offline_info_.function, offline_info_.model, "subnet0");

    cnrtGetInputDataSize(&offline_info_.inputSizeArray, &offline_info_.inputNum, offline_info_.function);
    cnrtGetOutputDataSize(&offline_info_.outputSizeArray, &offline_info_.outputNum, offline_info_.function);
    cnrtGetInputDataShape(&offline_info_.input_dims[0], &offline_info_.input_dims_num[0], 0, offline_info_.function);
    offline_info_.input_n = offline_info_.input_dims[0][0];
    offline_info_.input_h = offline_info_.input_dims[0][1];
    offline_info_.input_w = offline_info_.input_dims[0][2];
    offline_info_.input_c = offline_info_.input_dims[0][3];

    cnrtDestroy();

    return true;
}


bool InferClient ::parse_output_para() {
    for (int i = 0; i < offline_info_.outputNum; i++) {
        cnrtGetOutputDataShape(&offline_info_.output_dims[i], &offline_info_.output_dims_num[i], i, offline_info_.function);
        int n=1;
        int h=1;
        int w=1;
        int c=1;
        n = offline_info_.output_dims[i][0];
        if(offline_info_.output_dims_num[i]>1){
            h = offline_info_.output_dims[i][1];
        }
        else{
            offline_info_.output_dims[i][1] = 1;
        }
        if(offline_info_.output_dims_num[i]>2){
            w = offline_info_.output_dims[i][2];
        }
        else{
            offline_info_.output_dims[i][2] = 1;
        }
        if(offline_info_.output_dims_num[i]>3){
            c = offline_info_.output_dims[i][3];
        }
        else{
            offline_info_.output_dims[i][3] = 1;
        }
        offline_info_.output_n[i] = (unsigned int)n;
        offline_info_.output_h[i] = (unsigned int)h;
        offline_info_.output_w[i] = (unsigned int)w;
        offline_info_.output_c[i] = (unsigned int)c;
        offline_info_.out_count[i] = n*c*h*w;
    }
}

void InferClient ::infer_cur_data(void** mlu_input, void** mlu_output){
    InferWorkInfo* w = servers_[process_idx_];
    process_idx_ = ++process_idx_ % server_num_;
    processed_batch_ += offline_info_.input_n;
    std::pair<void**, void**> msg;
    msg.first = mlu_input;
    msg.second = mlu_output;
    w->qin->push(msg);
}

void InferClient ::fetch_result(){
    InferWorkInfo* w = servers_[fetch_idx_];
    int result;
    w->qout->pop(result);
    fetch_idx_ = ++fetch_idx_ % server_num_;
    fetched_batch_ = fetched_batch_ + offline_info_.output_n[0];
    return;
}

int InferClient :: get_processed_batch(){
    return processed_batch_;
}

int InferClient :: get_fetched_batch(){
    return fetched_batch_;
}

void* InferClient :: get_mlu_in_ptr(void* mlu_in){
    int offset = processed_batch_ * offline_info_.inputSizeArray[0]/offline_info_.input_n;
    void* p = (void*)((char*)mlu_in + offset);
    return p;
}

void* InferClient :: get_mlu_out_ptr(void* mlu_out){
    int offset = processed_batch_ * offline_info_.outputSizeArray[0]/offline_info_.input_n;
    void* p = (void*)((char*)mlu_out + offset);
    return p;
}

bool InferClient :: is_server_availed(){
    if(rest_ > 0){
        rest_ --;
        return true;
    }
    if(process_idx_ == fetch_idx_){
        return false;
    }else{
        return true;
    }
}

void InferClient :: process(void** mlu_in, void** mlu_out, int batch){
    init();
    while(true){
        if(get_fetched_batch() >= batch){
            return;
        }
        if(get_processed_batch() >= batch){
            fetch_result();
        }
        else{
            if(!is_server_availed()){
                fetch_result();
            }
            mlu_input_[process_idx_][0] = get_mlu_in_ptr(mlu_in[0]);
            mlu_output_[process_idx_][0] = get_mlu_out_ptr(mlu_out[0]);
            infer_cur_data(&mlu_input_[process_idx_][0], &mlu_output_[process_idx_][0]);
        }
    }
}
