//
// Created by jiapeiyuan on 4/24/21.
//

#include "server.h"


InferServer::InferServer(int server_id,
                                   const char* model_file,
                                   BlockQueue<std::pair<void**, void**>>* in,
                                   BlockQueue<int>* out){
    model_file_ = model_file;
    queue_in_ = in;
    queue_out_ = out;
    is_stop_ = false;
    server_id_ = server_id;

    new thread(&InferServer::start_work, this);
}

void InferServer::start_work(){
    init_worker();
    while(true){
        if(is_stop_){
            return;
        }
        std::pair<void**, void**> in_msg;
        queue_in_->pop(in_msg);
        void** in = in_msg.first;
        void** out = in_msg.second;
        start_infer(in, out);
        queue_out_->push(0);
    }
}

void InferServer::init_worker(){
    cnrtInit(0);
    init_device();
    parse_input_args();
    parse_output_args();
    init_function();
    init_ctxt();
}

bool InferServer::init_device() {
    unsigned dev_num;
    cnrtGetDeviceCount(&dev_num);
    if (dev_num == 0) {
        std::cout << "no device found" << std::endl;
        exit(-1);
    }
    cnrtDev_t dev;
    CNRT_CHECK(cnrtGetDeviceHandle(&dev, 0));
    CNRT_CHECK(cnrtSetCurrentDevice(dev));
    cnrtSetCurrentChannel((cnrtChannelType_t)(server_id_ % 4));

    return true;
}


bool InferServer ::init_ctxt() {
    cnrtCreateRuntimeContext(&rt_ctx_, function2, nullptr);
    cnrtSetRuntimeContextDeviceId(rt_ctx_, device_id_);
    cnrtSetRuntimeContextChannel(rt_ctx_, CNRT_CHANNEL_TYPE_NONE);
    cnrtInitRuntimeContext(rt_ctx_, nullptr);
    cnrtSetCurrentContextDevice(rt_ctx_);
}

bool InferServer::init_function(){
    cnrtCreateFunction(&function2);
    cnrtCopyFunction(&function2, function);
    cnrtCreateQueue(&queue_);
    cnrtCreateNotifier(&noti_start);
    cnrtCreateNotifier(&noti_end);
    return true;
}

void InferServer::parse_input_args(){
    cnrtLoadModel(&model, model_file_);
    int model_parallel;
    cnrtQueryModelParallelism(model, &model_parallel);

    string name = "subnet0";

    cnrtCreateFunction(&function);
    cnrtExtractFunction(&function, model, name.c_str());
    cnrtGetInputDataSize(&inputSizeArray_, &inputNum, function);
    cnrtGetOutputDataSize(&outputSizeArray_, &outputNum, function);

    cnrtGetInputDataShape(&input_dims, &input_dims_num, 0, function);
    input_n = input_dims[0];
    input_h = input_dims[1];
    input_w = input_dims[2];
    input_c = input_dims[3];

    data_parallel = 1;
    input_n *= data_parallel;
}


void InferServer ::parse_output_args() {
    for (int i = 0; i < outputNum; i++) {

        cnrtGetOutputDataShape(&output_dims[i], &output_dims_num[i], i, function);
        int n=1;
        int h=1;
        int w=1;
        int c=1;
        n = output_dims[i][0];
        if(output_dims_num[i]>1){
            h = output_dims[i][1];
        }
        else{
            output_dims[i][1] = 1;
        }
        if(output_dims_num[i]>2){
            w = output_dims[i][2];
        }
        else{
            output_dims[i][2] = 1;
        }
        if(output_dims_num[i]>3){
            c = output_dims[i][3];
        }
        else{
            output_dims[i][3] = 1;
        }
        output_n[i] = (unsigned int)n;
        output_h[i] = (unsigned int)h;
        output_w[i] = (unsigned int)w;
        output_c[i] = (unsigned int)c;
        out_count[i] = n*c*h*w;
    }
}

bool InferServer::start_infer(void** in, void** out){
    float event_time_use;
    void* param[inputNum + outputNum];
    for (int i = 0; i < inputNum; i++) {
        param[i] = in[i];
    }
    for (int i = 0; i < outputNum; i++) {
        param[inputNum + i] = out[i];
    }

    cnrtPlaceNotifier(noti_start, queue_);
    CNRT_CHECK(cnrtInvokeRuntimeContext(rt_ctx_, param, queue_, nullptr));
    cnrtPlaceNotifier(noti_end, queue_);

    if (cnrtSyncQueue(queue_) == CNRT_RET_SUCCESS) {
        cnrtNotifierDuration(noti_start, noti_end, &event_time_use);
        inferencer_time += event_time_use;
        std::cout << " execution time: " << event_time_use << std::endl;
    } else {
        std::cout << " SyncStream error" << std::endl;
    }

    return true;
}